
function backIsClose() {

  this.setupPrefs = setupPrefs ; // make public
  this.setup = setup ;
  this.updateState = updateState ;
  this.setClosePolicyGlobally = setClosePolicyGlobally ;
  this.shoutoutResponse = shoutoutResponse ;
  this.showPrefDialog = showPrefDialog ;
  this.doClose = doClose ;
  this.onClosing = onClosing ;

  const EXTENSIONNAME = "Back IS Close" ;
  const VERSION = "1.4" ;
  const PREFIX = "back-is-close" ;
  const PREFIXDASH = PREFIX + "-" ;
  const STRINGBUNDLE = PREFIXDASH + "strings" ;
  const CPMENUPREFIX = PREFIXDASH + "menu-cp" ;
  const BACKISCLOSEBUTTONID = PREFIXDASH + "back-button" ;
  const BICDISPLAYEDBROADCASTER = PREFIXDASH + "displayed-broadcaster" ;
  const BICKEYBROADCASTER = PREFIXDASH + "key-broadcaster" ;
  const BICGOMENUID = "back-is-close-goPopup" ;
  const IDFORANONYMOUSGOMENUITEM = PREFIXDASH + "go-back"
  // it's not, actually, a "back-is-close" item, but we use our prefix anyway, as the id must be unique
  const PREFPREFIX = PREFIX + "." ;
  const PREFCLOSEPOLICY = PREFPREFIX + "closePolicy" ;
  const PREFEASTEREGG = PREFPREFIX + "easterEgg" ;
  const PREFUSES = PREFPREFIX + "usesUntilPromptForNewVersion" ;
  const IGNORE_UPDATES_N_TIMES = 5 ;
  const BUTTONICONS = [ "1", "2", "2", "3" ] ; 
  const BUTTONICONATTRIBUTE = "wouldClose" ;
  const BUTONLABELSOFFSET = -1 ;
  const PREFWINDOWTYPE = PREFIX + ":preferences" ;
  const PREFWINDOWPATH = "chrome://back-is-close/content/back-is-close-prefs.xul" ;
  const PREFWINDOWSTYLE = "chrome" ;
  const LATESTVERSIONXMLFIELDS = [ "latestversion", "xpiurl", "confirmMessage" ] ;
  const SHOUTOUTPARAMETERNAMES = [ "version", "closePolicy", "easterEgg" ] ;
  const SHOUTOUTPROTOCOL = "GET" ;
  const SHOUTOUTURL = "http://www.diffenbach.org/back-is-close/latestversion.xml?" ;
  
  // attributes
  const ID = "id" ;
  const DISABLED = "disabled" ;
  const HIDDEN = "hidden" ;
  const ELEMENT = "element" ; 
  const LABEL = "label" ;
  const CHECKED = "checked" ;
  const TOOLTIP = "tooltiptext" ;
  
  // from browser.xul
  const GOMENUID = "goPopup" ;
  const BACKSPACECOMMANDID = "cmd_handleBackspace" ;
  const BACKCOMMANDID = "Browser:Back"; 
  const DISPLAYEDBACKIDS = [ "back-button", "context-back", IDFORANONYMOUSGOMENUITEM ] ;
  const BACKBUTTONID = DISPLAYEDBACKIDS[ 0 ]  ;
  
  // from Mozilla
  const MOZILLAPREFERNCESERVICE = "@mozilla.org/preferences-service;1"
  const MOZILLAWINDOWMEDIATORSERVICE = "@mozilla.org/appshell/window-mediator;1" ;
  const NAVIGATORBROWSER = "navigator:browser" ;
  
  // enum
  const closePolicyINACTIVE = 0 ; // extension not active: remove button, restore normal functionality
  const closePolicyTABMORETHANONE = 1 ; // close only if > 1 tabs
  const closePolicyWINDOWNOTOLDEST = 2 ; // close only if ( > 1 tabs || > 1 browser windows ) && window is not the oldest
  const closePolicyWINDOWMORETHANONE = 3 ; // close only if > 1 tabs || > 1 browser windows
  const closePolicyANYTABORWINDOW = 4 ; // close anything
  
  // enum
  const wouldCloseTAB = 1 ; //cp >= 1
  const wouldCloseWINDOW = 2 ; //cp >= 2
  const wouldCloseOLDEST =  3 ; //cp >=3
  const wouldCloseONLY = 4 ; //cp >=4 
  const wouldCloseERROR = 5 ;

  // Human-readable; should be read from a locale file, so can't be const, as StringBundle isn't available on initiation
  var CLOSEWINDOW ;
  var BUTTONLABELS ;
  var TTWINDOW ;
  var TOOLTIPS ;
  var EXCEPTION ;
  var EXCEPTIONMSGSETUP ;
  var EXCEPTIONMSGREADPREFS ;
  var EXCEPTIONMSGSAVEPREFS ;
  var EXCEPTIONMSGSHOUTOUT ;
  var EXCEPTIONMSGPREFWINDOW ;

  // state
  var stringBundle ;
  var originalUpdateBackForwardButtons = window.UpdateBackForwardButtons ;
  var originalBrowserBack = window.Browserback ;
  var backIsCloseIsHidden = true ;
  var closePolicy = closePolicyTABMORETHANONE ;
  var wouldClose = wouldCloseTAB ;
  var browserWindowCount = 1 ;
  var isOldestBrowserWindow = true ;
  
  var tries = 0 ;
  
  // utility functions

  function alertException( e, message ) {
    alert( EXTENSIONNAME + ": " + message + EXCEPTION + " " + e.name + " " + e.message + " " + e.toString() ) ;
  }
  
  // setup functions
  
  function setupStrings() {
    //var sb = document.getElementById( STRINGBUNDLE ) ;
    
    var sb = srGetStrBundle( "chrome://back-is-close/locale/back-is-close.properties" ) ;
    CLOSEWINDOW = sb.GetStringFromName( "back-is-close.label-window" ) ;
    BUTTONLABELS = [ sb.GetStringFromName( "back-is-close.label-tab" ), CLOSEWINDOW, CLOSEWINDOW, sb.GetStringFromName( "back-is-close.label-firefox" ) ] ;
    TTWINDOW = sb.GetStringFromName( "back-is-close.tooltip-window" ) ;
    TOOLTIPS = [ sb.GetStringFromName( "back-is-close.tooltip-tab" ) , TTWINDOW, TTWINDOW, sb.GetStringFromName( "back-is-close.tooltip-firefox" )  ] ;
    EXCEPTION = sb.GetStringFromName( "back-is-close.exception" ) ;
    EXCEPTIONMSGSETUP = sb.GetStringFromName( "back-is-close.exception-setup" ) ;
    EXCEPTIONMSGREADPREFS = sb.GetStringFromName( "back-is-close.exception-readprefs" ) ;
    EXCEPTIONMSGSAVEPREFS = sb.GetStringFromName( "back-is-close.exception-saveprefs" ) ;
    EXCEPTIONMSGSHOUTOUT = sb.GetStringFromName( "back-is-close.exception-shoutout" );
    EXCEPTIONMSGPREFWINDOW = sb.GetStringFromName( "back-is-close.exception-prefwindow" ) ;
    sb = null ;
  }
    
  // functions to correctly position bic menu item in "Go" Menu 
  function getAnonymousBackMenuItemInGoMenu() {
    var gm = document.getElementById( GOMENUID ) ;
    if( gm ) {
      var ea = gm.getElementsByAttribute( ELEMENT, BACKCOMMANDID ) ;
      if( ea && ea[ 0 ] && ea[ 0 ].parentNode ) {
        return ea[ 0 ].parentNode ;
      }
    }
    return null ;
  }
      
  function addIdToElement( e, id ) {
    e.setAttribute( ID, id ) ;
  }
  
  function repositionGoMenuItemBeforeElement( e ) {
    // because the "back" menuiten in the  "Go" meu has no id, we can't 
    // specfy the id in the "insertbefore" attribute in our matching "close" menuitem
    // so we have to manually move it here
  
    var c = document.getElementById( BICGOMENUID ) ;
    if( e && c ) {
      var p = e.parentNode ;
      p.insertBefore( c, e ) ;
    }
  }
  
  function fixUpGoMenu() {
    var e = getAnonymousBackMenuItemInGoMenu() ;
    addIdToElement( e, IDFORANONYMOUSGOMENUITEM ) ;
    repositionGoMenuItemBeforeElement( e ) ;
  }
  
  
  function addBackIsCloseButtonToToolbar() {
    var bicButton = document.getElementById( BACKISCLOSEBUTTONID ) ;
    // if it's found, we don't care where; if the user wants to be perverse and move it, so be it
    if( bicButton != null ) {
      bicButton.parentNode.removeChild( bicButton ) ;
      bicButton = null ;
    }
    if( bicButton == null ) {
      var backButton = document.getElementById( BACKBUTTONID ) ;
      var t ;
      if( backButton && ( t = backButton.parentNode ) ) {
        bicButton = t.insertItem( BACKISCLOSEBUTTONID, backButton, null, true ) ;
      }
    }
  }

  function setupFunctions() {
    originalUpdateBackForwardButtons = window.UpdateBackForwardButtons ;
    originalBrowserBack = window.BrowserBack ;
    window.UpdateBackForwardButtons = bicUpdateBackForwardButtons ;
    window.BrowserBack = bicBrowserBack ;
  }


  // preference service functions

  function getPreferencesService() {
    return Components.classes[ MOZILLAPREFERNCESERVICE ]
      .getService(Components.interfaces.nsIPrefBranch) ;
  }
  
  function savePrefs( policy, easter ) {
    try {
      var prefs = getPreferencesService() ;
      prefs.setIntPref( PREFCLOSEPOLICY, policy ) ;
      if( easter ) {
        prefs.setIntPref( PREFEASTEREGG, easter ) ;
      }
    } catch( e ) {
      alertException( e, EXCEPTIONMSGSAVEPREFS ) ;
    }
  }
  
  function setupPrefs() {
    // default is already set, no need to set it
    try {
      var prefs = getPreferencesService() ;
      var policy = prefs.getIntPref( PREFCLOSEPOLICY ) ;
    } catch( e ) {
      //alertException( e, EXCEPTIONMSGREADPREFS) ;
      // in case it's because the the pref doesn't exist, try to save the pref
      //savePrefs( closePolicy ) ;
      //policy = closePolicyTABMORETHANONE ;
    }
    updateAllBrowserWindows( policy ) ;
    return prefs ;
  }


  // shoutout functions
  
  function parseLatestVersionAndUrl( rx ) {
    var tns = LATESTVERSIONXMLFIELDS ;
    var a = new Array() ; 
    var i = 0 ;
    for( ; i < tns.length ; ++i ) {
      var h = rx.getElementsByTagName( tns[ i ] ) ;
      if( h && h[ 0 ] && h[ 0 ].firstChild ) {
        a[ tns[ i ] ] = h[ 0 ].firstChild.data ;
      } else break ;
    }
    return i == tns.length ? a : null ;
  }
  
  function downloadNewVersion( xpiurl ) {
    window.open( xpiurl, PREFIX ) ;
  }
  
  function shoutoutResponse( rx, prefs ) {
    try {
      var a = parseLatestVersionAndUrl( rx ) ;
      if( a && a.latestversion != VERSION ) {
        var nv = prefs.getIntPref( PREFUSES ) ;
        if( nv < 1 ) {
          if( confirm( a.confirmMessage ) ) {
            downloadNewVersion( a.xpiurl ) ;
            nv = 0 ;
          } else {
            nv = IGNORE_UPDATES_N_TIMES ;
          }
        } else {
          nv -= 1 ;
        }
        prefs.setIntPref( PREFUSES, nv ) ;
      }
    } catch( e ) {
      alertException( e, EXCEPTIONMSGSHOUTOUT  ) ;
    }
  }
  
  function shoutout( version, policy, prefs ) {
    req = new XMLHttpRequest();
    var data = [ version, policy, prefs.getIntPref( PREFEASTEREGG ) ] ;
    var names = SHOUTOUTPARAMETERNAMES ;
    var parms = "" ;
    for( var i = 0 ; i < names.length ; ++i ) {
      parms += escape( names[ i ] ) + "=" + escape( data[ i ] ) + "&" 
    }
    parms = parms.substr( 0, parms.length - 1 ) ;
    req.open( SHOUTOUTPROTOCOL, SHOUTOUTURL + parms, true );
    req.onreadystatechange = function( aEvt ) {
      if( req.readyState == 4 ) {
        if( !req.status || req.status == 200 ) {
          gBackIsClose.shoutoutResponse( req.responseXML, prefs ) ;
        }
      }
    } ;
    req.send(null); 
  }


  // public function called to bootstrap; calling it on creation fails because browser is not yet created
  
  function setup() {
  var w = 0 ;
    try {
      setupStrings() ;
      ++w ;
      fixUpGoMenu() ;
      ++w ;
      addBackIsCloseButtonToToolbar() ;
      ++w ;
      setupFunctions() ;
      ++w ;
      var prefs = setupPrefs() ; // counter-intuitively, this should be (nearly) last because it has side-effects
      ++w ;
//      if( isOldestBrowserWindow ) 
//        shoutout( VERSION, closePolicy, prefs ) ;
//      ++w ;
    } catch( e ) {
      if( tries++ == 0 ) {
        alertException( e, EXCEPTIONMSGSETUP + " will retry " + w + " " ) ;
        setup()
      } else {
        alertException( e, EXCEPTIONMSGSETUP + " " + w + " " ) ;
      }
    }
  }
  
  // end setup functions
  
  
  // functions to keep state consistent over all windows
  
  function getWindowMediator() {
    return Components.classes[ MOZILLAWINDOWMEDIATORSERVICE ].getService(Components.interfaces.nsIWindowMediator) ;
  }
  
  function getBrowserWindowEnumerator( windowType ) {
    return getWindowMediator().getEnumerator( windowType ) ;
  }          
  
  function getMostRecentWindow( windowType ) {
    return getWindowMediator().getMostRecentWindow( windowType ) ;
  }
  
  function updateAllBrowserWindows( policy ) {
    var en = getBrowserWindowEnumerator( NAVIGATORBROWSER ) ;
    var a = new Array() ;
    while( en.hasMoreElements() ) {
      a.push( en.getNext() ) ;
    }
    var c = a.length ;
    for( var i = 0 ; i < c ; ++i ) {
      a[ i ].gBackIsClose.updateState( c, i == 0, policy ) ;
    }
  }
  
  
  // public function called on all windows by updateAllBrowserWindows

  function updateState( count, oldest, newPolicy ) {
    if( newPolicy != closePolicy ) {
      //alert( "|" + CPMENUPREFIX + closePolicy + "|" + CPMENUPREFIX + newPolicy + "|" ) ;
      document.getElementById( CPMENUPREFIX + closePolicy ).removeAttribute( CHECKED ) ; 
      document.getElementById( CPMENUPREFIX + newPolicy ).setAttribute( CHECKED, true ) ; 
      closePolicy = newPolicy ;
    }
    browserWindowCount = count ;
    isOldestBrowserWindow = oldest ;
    try {
      window.UpdateBackForwardButtons() ;
    } catch( e ) {
    }
  }   
  
  
  // public function called by our easter egg menu
  
  function showPrefDialog() {
  // thanks to ausdilecce and TheOneKEA  
    try {
      var wm = getMostRecentWindow( PREFWINDOWTYPE );
      wm ? wm.focus() : window.openDialog( PREFWINDOWPATH, "", PREFWINDOWSTYLE, null ) ;
    } catch( e ) {
      alertException( e, EXCEPTIONMSGPREFWINDOW ) ;
    }
  }




  // public function called by our preference setting easter-egg menu 
  
  function setClosePolicyGlobally( policy ) {  
    //setTimeout( ( function() {
      savePrefs( policy, 1 ) ;
      updateAllBrowserWindows( policy ) ;
      var w = getMostRecentWindow( PREFWINDOWTYPE ) ;
      if( w ) 
        w.updatePolicy( policy ) ;
    //} ), 1 ) ;
  }


  // public function to call other windows when this window closes

  function onClosing() {
    updateAllBrowserWindows( closePolicy )  ;
  }

  // GUI mutators
  
  function elementArrayFromIdArray( ia ) {
    ret = new Array() 
    for( var i = 0 ; i < ia.length ; ++i ) 
      ret.push( document.getElementById( ia[ i ] ) ) ;
    return ret ;
  }
  
  function applySetAttributeToElementArray( ea, a, v ) {
    for( var i = 0 ; i < ea.length ; ++i ) 
      ea[ i ].setAttribute( a, v ) ;
  }
  
  function applyRemoveAttributeToElementArray( ea, a ) {
    for( var i = 0 ; i < ea.length ; ++i ) 
      ea[ i ].removeAttribute( a ) ;
  }
  
  function setOrRemoveAttributeOnIdArray( set, ids, attribute, value ) {
  // we could also use function pointers, but I fear time-overhead
    var els = elementArrayFromIdArray( ids ) ;
    if( set ) {
      applySetAttributeToElementArray( els, attribute, value ) ;
    } else {
      applyRemoveAttributeToElementArray( els, attribute ) ;
    }
  }
  
  function setStyleAttributeOnIdArray( ids, attribute, value ) {
    var els = elementArrayFromIdArray( ids ) ;
    for( var i = 0 ; i < ids.length ; ++i ) 
      els[ i ].style[ attribute ] = value ;
    }
  
  function hideAndDisableBack( hide ) {
    //setOrRemoveAttributeOnIdArray( hide, BACKCOMMANDIDS, DISABLED, true ) ;
    var bsc = document.getElementById( BACKSPACECOMMANDID ) ;
    if( hide ) 
      bsc.setAttribute( DISABLED, true ) ;
    else 
      bsc.removeAttribute( DISABLED ) ;
      
    //setOrRemoveAttributeOnIdArray( hide, DISPLAYEDBACKIDS, HIDDEN, true ) ;
    setStyleAttributeOnIdArray( DISPLAYEDBACKIDS, "display", hide ? "none" : "" ) ;
  }
  
  function hideAndDisableClosed( hide, displaybroadcaster ) {
    var keybroadcaster = document.getElementById( BICKEYBROADCASTER )
    if( hide ) {
      keybroadcaster.setAttribute( DISABLED, true ) ;
      displaybroadcaster.setAttribute( HIDDEN, true ) ;
    } else {
      keybroadcaster.removeAttribute( DISABLED ) ;
      displaybroadcaster.removeAttribute( HIDDEN ) ;
    }
  }
  
  function updateCloseLook( newWouldClose, displaybroadcaster ) {
    var offset = newWouldClose + BUTONLABELSOFFSET ;
    displaybroadcaster.setAttribute( LABEL, BUTTONLABELS[ offset ] ) ;
    displaybroadcaster.setAttribute( TOOLTIP, TOOLTIPS[ offset ] ) ; 
    displaybroadcaster.setAttribute( BUTTONICONATTRIBUTE, BUTTONICONS[ offset ] ) ;
    wouldClose = newWouldClose ;
  }
  
  function getBicDisplayBroadcaster() {
    return document.getElementById( BICDISPLAYEDBROADCASTER ) ;
  }
  
  function switchGUIToBIC( currentWouldClose ) {
    // this function rather kludgily takes either false or a number > 0
    // we have to cast back to a bool, because the 2nd "if" compares two bools for equality
    var displaybroadcaster = null ;
    var displayBackIsClose = ( currentWouldClose > 0 ) ;
    
    if( displayBackIsClose && currentWouldClose != wouldClose ) {
      updateCloseLook( currentWouldClose, ( displaybroadcaster = getBicDisplayBroadcaster() ) ) ;
    }
    
    if( displayBackIsClose == backIsCloseIsHidden ) {
      hideAndDisableBack( displayBackIsClose ) ;
      hideAndDisableClosed( !displayBackIsClose, displaybroadcaster ? displaybroadcaster : getBicDisplayBroadcaster() ) ;
      backIsCloseIsHidden = !displayBackIsClose ;
    }
  }
  
  
  // Policy appliers
  
  function moreThanOneTab() {
    var b = getBrowser() ;
    return b && b.browsers && b.browsers.length > 1 ;
  }
  
  function wouldCloseWhat() {
    try {
      return moreThanOneTab() ? wouldCloseTAB
        : ( browserWindowCount < 2 ? wouldCloseONLY
          : ( isOldestBrowserWindow ? wouldCloseOLDEST 
            : wouldCloseWINDOW ) ) ;
    } catch( e ) {
      return wouldCloseERROR ;
    }
  }
  
  
  // Overridden browser.js window function
  
  function bicUpdateBackForwardButtons() {
    originalUpdateBackForwardButtons() ;
    var currentWouldClose ;
    switchGUIToBIC( 
      document.getElementById( BACKCOMMANDID ).hasAttribute( DISABLED ) 
      && ( currentWouldClose = wouldCloseWhat() ) <= closePolicy && currentWouldClose )  ;
      // This takes advantage of a javscript mis-feature, that logical operations "return" the last operand evaluated
      // rather than just true or false, and the C-ish feature that logical operators are "sequence points" (enforce order of evaluation)
  }
  
  // Overridden browser.js function 
  function bicBrowserBack( aEvent, aIgnoreAlt ) {
    if( backIsCloseIsHidden ) {
      originalBrowserBack( aEvent, aIgnoreAlt ) ;
    } else {
      doClose() ;
    }
  }
      
  // function called by our GUI elements

  function doClose() {
    wouldClose == wouldCloseTAB ? getBrowser().removeCurrentTab() : window.close() ;
  }
  

    
  //setup() ;
} // end backIsClose function
  
var gBackIsClose = new backIsClose() ;
  
function tpdomd( e ) {
  alert( "buttons " + e.button ) ;
}

window.addEventListener( "onmousedown", tpdomd, false ) ; 

window.addEventListener( "load", gBackIsClose.setup, false ) ; 
window.addEventListener( "unload", gBackIsClose.onClosing, false) ; 
  
